import pendulum
from airflow.decorators import dag, task
from airflow.models.baseoperator import chain
from airflow.operators.bash import BashOperator
from airflow.sensors.filesystem import FileSensor

default_args = {
    "owner": "mr.hasan",
    "email": ["tester@pesnik.com"],
    "email_on_failure": False,
    "email_on_retry": False,
}


@dag(
    default_args=default_args,
    schedule=None,
    start_date=pendulum.today("Asia/Dhaka").add(days=0),
    tags=["TEST"],
    on_failure_callback=lambda x: print("Failed"),
    on_success_callback=lambda x: print("Success"),
    params={
        "date_value": pendulum.today("Asia/Dhaka").add(days=0).strftime("%Y-%m-%d")
    },
    max_active_runs=1,
    catchup=False,
)
def dag_tester():
    @task
    def start():
        print("Job Started @10")

    task_start = start()

    task_sensor = FileSensor(
        task_id="file_sensor",
        filepath=FILE_PATH,
        poke_interval=60,
        timeout=600,
        sql="SELECT 1;",
    )

    @task.bash
    def cmd_task(message):
        return f"echo {message}"

    @task.bash
    def cmd_task_sh(message):
        return f"/home/laboratory/dagsonar/playground/greeting_bot.sh {message}"

    message = "Hello World Glitch"
    task_cmd = cmd_task(message=message)

    task_bash_op = BashOperator(
        task_id="bash_op",
        bash_command="/home/laboratory/dagsonar/playground/greeting_bot.sh ",
    )

    @task
    def end():
        print(END_MESSAGE)
        caller()
        print("Job Completed")

    chain(
        task_start,
        task_sensor,
        task_cmd,
        cmd_task_sh("Cluster hop, on the top"),
        task_bash_op,
        end(),
    )
    # task_start >> task_sensor >> task_cmd >> end()


END_MESSAGE = "End of DAG 1"
FILE_PATH = "/usr/local/airflow/dags/initiator"


def caller():
    print("Has been called!")


dag_tester()



from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Set


@dataclass
class DagConfig:
    path: Path
    tasks: List[str]


@dataclass
class ShellScriptReference:
    path: Path
    content: str


@dataclass
class ExprReference:
    name: str
    content: str


@dataclass
class TaskReference:
    task_id: Optional[str] = None
    task_type: Optional[str] = None
    content: Optional[str] = None
    shell_scripts: List[ShellScriptReference] = field(
        default_factory=list
    )  # [(path, content)]
    external_variables: List[ExprReference] = field(
        default_factory=list
    )  # [(name, content)]
    called_functions: List[ExprReference] = field(
        default_factory=list
    )  # [(name, content)]
    hash: Optional[str] = None
    last_modified: Optional[str] = None


@dataclass
class DagReference:
    dag_id: str
    task_history: List[TaskReference]



import ast
import json
from pathlib import Path
from typing import Dict, List, Set, Union
from datetime import datetime

from dagsonar import (DagConfig, DagReference, ExprReference, Parser,
                      ReferenceEncoder, ShellScriptReference, TaskReference,
                      compute_hash)


class TaskTracker:
    """Main class for tracking DAG task changes."""

    def __init__(self, history_file: Path = Path("task_history.json")):
        self.history_file = history_file
        self.dag_history = self._load_history()

    def _load_history(self):
        try:
            if self.history_file.exists():
                with open(self.history_file, "r") as f:
                    return json.load(f)
        except json.JSONDecodeError as e:
            print(f"Error loading history file: {e}")

    def save_history(self, reference: List[Dict[str, str | DagReference]]):
        with open(self.history_file, "w") as f:
            json.dump(reference, f, indent=2, cls=ReferenceEncoder)

    def start_monitoring(self):
        """Start monitoring DAG tasks for changes."""
        pass

    def track_tasks(
        self, dag_configs: Dict[str, DagConfig]
    ) -> List[Dict[str, str | DagReference]]:
        dag_references: List[Dict[str, str | DagReference]] = []
        for dag, config in dag_configs.items():
            if not config.path.exists():
                print(f"Warning: Dag file not found: {config.path}")
                continue
            else:
                history = list()
                parser = Parser(config.path)
                tasks = parser.get_tasks(config.tasks)
                for task in tasks:
                    tracks = self._process_task(task, parser)
                    history.append(tracks)
                dag_references.append(
                    {
                        "dag_id": dag,
                        "reference": DagReference(dag_id=dag, task_history=history),
                    }
                )

        return dag_references

    def check_for_changes(
        self, new_reference: List[Dict[str, Union[str, DagReference]]]
    ) -> List[Dict[str, str]]:
        """
        Returns a list of task IDs whose hash values have changed.
        """
        result = []

        for dag in new_reference:
            changed_tasks = []
            reference: DagReference = dag["reference"]

            if not isinstance(reference, DagReference):
                continue

            dag_id = reference.dag_id
            new_hashes = {task.task_id: task.hash for task in reference.task_history}
            print(dag_id, new_hashes)

            history_reference = None
            if self.dag_history is None:
                continue

            for dag in self.dag_history:
                ref = dag.get("reference")
                if ref["dag_id"] == dag_id:
                    history_reference = ref

            if history_reference is None:
                continue

            old_hashes = {
                task["task_id"]: task["hash"]
                for task in history_reference["task_history"]
            }

            for task_id, new_hash in new_hashes.items():
                if task_id not in old_hashes or old_hashes[task_id] != new_hash:
                    changed_tasks.append(task_id)

            result.append({"dag": reference.dag_id, "tasks": changed_tasks})

        for changed_tasks_list in result:
            if len(changed_tasks_list) <= 0:
                continue

            for dag in new_reference:
                reference_blocks = dag['reference'].task_history
                for ref in reference_blocks:
                    if ref.task_id in changed_tasks_list['tasks']:
                        ref.last_modified = str(datetime.now())
                
        return result, new_reference

    def _process_task(
        self, task: ast.FunctionDef | ast.Call, parser: Parser
    ) -> TaskReference:
        reference = TaskReference()
        if isinstance(task, ast.FunctionDef):
            reference.task_id = task.name
            reference.content = ast.dump(task)

            decorators = task.decorator_list
            decorated_bash = False
            for decorator in decorators:
                if isinstance(decorator, ast.Attribute):
                    decorated_bash = decorator.attr == "bash"

            def traverse(head, intend=0):
                # print("\t->" * intend + (ast.dump(head)))

                # Check .sh file to track
                if decorated_bash and isinstance(head, ast.Constant):
                    files = [
                        file for file in head.value.split(" ") if file.endswith(".sh")
                    ]
                    for file in files:
                        with open(file, "r") as f:
                            content = f.read()
                            reference.shell_scripts.append(
                                ShellScriptReference(
                                    path=Path(file),
                                    content=ast.dump(ast.Constant(content)),
                                )
                            )

                if isinstance(head, ast.Name):
                    var = parser.find_variable_reference(variable=head)
                    if var.value is not None:
                        reference.external_variables.append(
                            ExprReference(name=head.id, content=ast.dump(var))
                        )

                if isinstance(head, ast.Call):
                    if isinstance(head.func, ast.Name):
                        fn = parser.find_function_reference(fn=head.func)
                        if fn is not None:
                            reference.called_functions.append(
                                ExprReference(name=fn.name, content=ast.dump(fn))
                            )

                for child in ast.iter_child_nodes(head):
                    traverse(child, intend + 1)

            traverse(task)
        elif isinstance(task, ast.Call):
            bash_operator = (
                isinstance(task.func, ast.Name) and task.func.id == "BashOperator"
            )

            task_id = next(
                (
                    keyword.value
                    for keyword in task.keywords
                    if keyword.arg == "task_id"
                ),
                ast.Constant,
            )
            reference.task_id = task_id.value
            reference.content = ast.dump(task)

            for arg in task.args:
                pass
            for keyword in task.keywords:
                if isinstance(keyword.value, ast.Name):
                    ref = parser.find_variable_reference(keyword.value)
                    keyword.value = ref
                    reference.external_variables.append(
                        ExprReference(
                            name=keyword.value.value, content=ast.dump(keyword)
                        )
                    )

                if (
                    bash_operator
                    and keyword.arg == "bash_command"
                    and isinstance(keyword.value, ast.Constant)
                ):
                    files = [
                        file
                        for file in keyword.value.value.split(" ")
                        if file.endswith(".sh")
                    ]
                    for file in files:
                        with open(file, "r") as f:
                            content = f.read()
                            reference.shell_scripts.append(
                                ShellScriptReference(
                                    path=Path(file),
                                    content=ast.dump(ast.Constant(content)),
                                )
                            )
        reference.hash = compute_hash(reference)
        return reference



import hashlib
import json


def compute_hash(reference) -> str:
    """Compute a SHA-256 hash of a given object."""
    reference_dict = reference.__dict__.copy()
    reference_dict.pop("hash", None)
    json_str = json.dumps(reference_dict, default=str, sort_keys=True)
    return hashlib.sha256(json_str.encode()).hexdigest()



import json
from dataclasses import asdict
from pathlib import Path

from dagsonar.models import (DagReference, ExprReference, ShellScriptReference,
                             TaskReference)


class ReferenceEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(
            o, (DagReference, TaskReference, ShellScriptReference, ExprReference)
        ):
            return asdict(o)

        if isinstance(o, Path):
            return str(o)

        return super().default(o)



import ast
from typing import List


class Parser:
    def __init__(self, file_loc):
        self.file_loc = file_loc
        with open(file_loc, "r") as f:
            self.tree = ast.parse(f.read())
            # debug(self.tree)

    def get_tasks(self, task_ids: List[str]) -> List[ast.FunctionDef | ast.Call]:
        tasks = []
        for node in ast.walk(self.tree):
            if isinstance(node, ast.Assign):
                for target in node.targets:
                    if (
                        isinstance(target, ast.Name)
                        and isinstance(node.value, ast.Call)
                        and target.id in task_ids
                    ):
                        if isinstance(node.value.func, ast.Name):
                            task = self.check_taskflow_operator(node.value.func.id)
                            if task is not None:
                                tasks.append(task)
                            else:
                                tasks.append(node.value)
            elif isinstance(node, ast.FunctionDef) and node.name in task_ids:
                tasks.append(node)

        return tasks

    def check_taskflow_operator(self, id: str):
        for node in ast.walk(self.tree):
            if (
                isinstance(node, ast.FunctionDef)
                and node.decorator_list is not None
                and node.name == id
            ):
                return node
        return None

    def find_variable_reference(self, variable: ast.Name) -> ast.Constant:
        for node in ast.walk(self.tree):
            if isinstance(node, ast.Assign):
                for target in node.targets:
                    if isinstance(target, ast.Name) and target.id == variable.id:
                        if isinstance(node.value, ast.Constant):
                            return node.value

        return ast.Constant(value=None)

    def find_function_reference(self, fn: ast.Name) -> ast.FunctionDef | None:
        for node in ast.walk(self.tree):
            if isinstance(node, ast.FunctionDef):
                if node.name == fn.id:
                    return node
        return None



# SPDX-FileCopyrightText: 2025-present Md. Rakibul Hasan <mr.hasan@banglalink.net>
#
# SPDX-License-Identifier: MIT
from dagsonar.__about__ import __version__
from dagsonar.encoder import ReferenceEncoder
from dagsonar.hash import compute_hash
from dagsonar.models import (DagConfig, DagReference, ExprReference,
                             ShellScriptReference, TaskReference)
from dagsonar.parser import Parser
from dagsonar.tracker import TaskTracker

__all__ = [
    "Parser",
    "TaskTracker",
    "DagConfig",
    "__version__",
    "TaskReference",
    "DagReference",
    "ShellScriptReference",
    "ExprReference",
    "compute_hash",
    "ReferenceEncoder",
]



from pathlib import Path

from dagsonar import DagConfig, TaskTracker


def test_task_tracker():
    tracker = TaskTracker()

    config = {
        "tester": DagConfig(
            path=Path("/Users/r_hasan/Development/dagsonar/playground/dag_tester.py"),
            tasks=["end", "start", "cmd_task_sh", "task_bash_op"],
        )
    }
    new_reference = tracker.track_tasks(config)
    change_detected = tracker.check_for_changes(new_reference)
    tracker.save_history(new_reference)
    print(change_detected)
